// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef GRAPHICS_SYSTEM_H
#define GRAPHICS_SYSTEM_H

#include <vector>
#include <string>
#include "mtxlib.h"

/*!
    \file graphicssystem.h
    \brief File contains abstract interface for a generic Graphics System.
*/


#define VTX_XYZ         0x00000001
#define VTX_NORMAL      0x00000002
#define VTX_COLOUR      0x00000004
#define VTX_UV1         0x00000008
#define VTX_UV2         0x00000010
#define VTX_UV3         0x00000020

//! a structure describing a colour.
class COLOUR
{
public:
    COLOUR() {}
    ~COLOUR() {}
    // Constructor with initializing values for r, g, b and optinally a
    COLOUR(float inR, float inG, float inB, float inA = 1.0f): r(inR), g(inG), b(inB), a(inA) {}
    // Constructor with initializing COLOUR
    COLOUR(const COLOUR &c): r(c.r), g(c.g), b(c.b), a(c.a) {}

    float r;    //!< Red component,   0.0f Black        1.0f Red
    float g;    //!< Green component, 0.0f Black        1.0f Green
    float b;    //!< Blue component,  0.0f Black        1.0f Blue
    float a;    //!< Alpha component, 0.0f Transparent  1.0f Opaque
};

struct VERTEX_XYZ
{
    vector3 point;
};

struct VERTEX_XYZ_COLOUR_UV1
{
    vector3 point;
    COLOUR colour;
    float u, v;
};

struct VERTEX_XYZ_NORMAL
{
    vector3 point;
    vector3 normal;
};

struct VERTEX_XYZ_NORMAL_UV1
{
    vector3 point;
    vector3 normal;
    float u, v;
};

struct VERTEX_XYZ_NORMAL_COLOUR_UV1_UV2
{
    vector3 point;
    vector3 normal;
    COLOUR colour;
    float u1, v1, u2, v2;
};

struct VERTEX_XYZ_UV1
{
    vector3 point;
    float u, v;
};

struct VERTEX_XYZ_UV1_UV2
{
    vector3 point;
    float u1, v1, u2, v2;
};


/*!
    An enum that is passed to DrawTriangles to indicate the way the points
    are organised
*/
enum TRIANGLES_TYPE
{
    TRIANGLE_LIST,
    TRIANGLE_STRIP,
    TRIANGLE_FAN
};

/*!
    An enum of types that are passed to SetTransformationMatrix to tell it
    which transformation matrix to change.
*/
enum TRANSFORMATION_TYPE
{
    OBJECT_TO_WORLD_TRANSFORM,
    VIEW_TRANSFORM,
    PROJECTION_TRANSFORM
};

//! A material used when lighting a surface.
struct MATERIAL
{
    COLOUR ambient;
    COLOUR diffuse;
    COLOUR specular;
    COLOUR emissive;
    float shiny;        // Sharpness of specular highlights. 1.0 shiny.
                        // 0.0 not shiny.
};

//! Type of light
enum LIGHTTYPE
{
    POINT_LIGHT,        //!< Point light. Radiates light in all directions.
    SPOT_LIGHT,         //!< Spot light. Cone shaped.
    DIRECTIONAL_LIGHT   //!< Like a point light infinitely far away.
};

//! A structure that describes a light
struct LIGHT
{
    LIGHTTYPE   type;       //!< Type of light e.g. POINT_LIGHT
    COLOUR      diffuse;    //!< Diffuse colour of light
    COLOUR      specular;   //!< Specular colour
    COLOUR      ambient;    //!< Ambient colour
    vector3     position;   //!< Position. Ignored if DIRECTIONAL_LIGHT
    vector3     direction;  //!< Direction. Ignored if POINT_LIGHT
    float       range;      //!< Range of light's influence. Ignored if DIRECTIONAL_LIGHT
    float       falloff;    //!< Decrease in illumination between 2 cones.
    float       attenuation0;
    float       attenuation1;
    float       attenuation2;
    float       theta;      //!< Angle in radians of SPOT_LIGHTs inner cone
    float       phi;        //!< Angle in radians of SPOT_LIGHTs outer cone
};

//! Texture wrapping modes
enum TEXTURE_MODE
{
    WRAP_TEXTURE,
    CLAMP_TEXTURE
};

enum TEXTURE_BLEND_MODE
{
    NORMAL,
    DOT3_BUMP,
    DETAIL_TEXTURE
};

//! Culling modes
enum cull_mode
{
    CULL_CW,        // Cull clockwise triangles
    CULL_CCW,       // Cull counter-clockwise triangles
    CULL_NONE
};

enum depth_buffer_mode
{
    READ_ONLY,      // Useful when drawing transparent objects
    READ_AND_WRITE, // Normal operation
    IGNORE_DEPTH    // Useful for drawing the Heads Up Display (HUD).
};

enum blend_mode
{
    SRC_ALPHA_ONE,                  // For explosions or flares or flames
    SRC_ALPHA_ONE_MINUS_SRC_ALPHA,  // For smoke or dirt
    ONE_ONE,                        // Additive blending
    ONE_ONE_MINUS_SRC_COLOR,
    BLEND_DISABLE                   // For everything else
};

enum shadow_mode
{
    Z_FAIL_SHADOWS,
    Z_PASS_SHADOWS
};

//! A handle which indirectly references a vertex buffer.
typedef int VERTEX_BUFFER_HANDLE;

//! A handle which indirectly references an index buffer.
typedef int INDEX_BUFFER_HANDLE;

//! A handle which indirectly references a texture map.
typedef int TEXTURE_HANDLE;

//! Abstract interface for a Graphics Substem.
class IGraphicsSystem
{
public:
    //! Initialises the graphics system
    IGraphicsSystem() {};

    //! Deinitialises the graphics system
    virtual ~IGraphicsSystem() {};

    //! Clears the screen
    virtual void Clear(COLOUR colour) = 0;

    //! Begins the scene.
    virtual void Begin() = 0;

    //! Ends the scene.
    virtual void End() = 0;

    //! Displays the backbuffer onto the screen.
    virtual void Display() = 0;

    virtual VERTEX_BUFFER_HANDLE CreateVertexList(int vertex_format,
        float* vertex_list, int number_of_vertices) = 0;

    //! Deletes a vertex buffer
    virtual void DeleteVertexList(VERTEX_BUFFER_HANDLE handle) = 0;

    //! Creates an index buffer. An array of indices.
    virtual INDEX_BUFFER_HANDLE CreateIndexBuffer(std::vector<unsigned short> index_list) = 0;

    //! Deletes an index buffer.
    virtual void DeleteIndexBuffer(INDEX_BUFFER_HANDLE handle) = 0;

    //! Creates a texture map from a PNG file
    virtual TEXTURE_HANDLE CreateTexture(std::string texture_filename,
        bool generate_mipmaps, bool alpha = false) = 0;

    //! Creates a texture map from a memory buffer.
    /*!
        buffer should be an already allocated memory buffer of size width * height * 3 bytes.
        Should be 24bits depth. RGBRGBRGB... order.
    */
    virtual TEXTURE_HANDLE CreateTexture(unsigned char *buffer, int width, int height,
        bool generate_mipmaps) = 0;

    //! Deletes a texture
    virtual void DeleteTexture(TEXTURE_HANDLE handle) = 0;

    //! Sets the texture for multitexturing stage.
    virtual void SetTexture(int stage, TEXTURE_HANDLE handle) = 0;

    //! Sets the texture blend mode.
    virtual void SetMultiTextureMode(TEXTURE_BLEND_MODE mode) = 0;

    //! Sets the texture wrapping mode
    virtual void SetTextureWrappingMode(int stage, TEXTURE_MODE mode) = 0;

    //! Sets the culling mode
    virtual void SetCullMode(cull_mode cull) = 0;

    //! Sets which material is used for lighting.
    virtual void SetMaterial(MATERIAL material) = 0;

    //! Draws triangles from a vertex buffer.
    virtual void DrawTriangles(TRIANGLES_TYPE type, VERTEX_BUFFER_HANDLE vhandle, int vertex_count) = 0;

    //! Draws triangles from a vertex buffer and index buffer.
    virtual void DrawIndexedTriangles(TRIANGLES_TYPE type, VERTEX_BUFFER_HANDLE vhandle, int vertex_count,
        INDEX_BUFFER_HANDLE ihandle, int index_count) = 0;

    //! Draws triangles from a vertex array
    virtual void DrawTrianglesUP(TRIANGLES_TYPE type, int vertex_format, float *vertex_array,
        int vertex_count) = 0;

    //! Draws triangles from a vertex array and index array
    virtual void DrawIndexedTrianglesUP(TRIANGLES_TYPE type, int vertex_format, float *vertex_array,
        int vertex_count, unsigned short *index_array, int index_count) = 0;

    //! Sets a light
    virtual void SetLight(int light_number, LIGHT light) = 0;

    //! Sets the colour of ambient light in the scene
    virtual void SetAmbientLight(COLOUR colour) = 0;

    //! Enables/Disables lighting.
    virtual void EnableLighting(bool enable) = 0;

    //! Enables or disables a light
    virtual void EnableLight(int light_number, bool enable) = 0;

    virtual void SetColour(COLOUR colour) = 0;

    //! Sets a transformation matrix.
    virtual void SetTransformationMatrix(TRANSFORMATION_TYPE type, const matrix44&) = 0;

    //! Returns a transformation matrix.
    virtual matrix44 GetTransformationMatrix(TRANSFORMATION_TYPE type) = 0;

    //! Sets the fog parameters
    virtual void FogParams(float density, COLOUR colour) = 0;

    //! Enables or disables fog
    virtual void EnableFog(bool enable) = 0;

    //! Sets the blend mode
    virtual void SetBlendMode(blend_mode blend) = 0;

    //! Sets the depth buffer mode
    virtual void SetDepthBufferMode(depth_buffer_mode depth) = 0;
    
    //! Enables or disables wireframe mode
    virtual void EnableWireframe(bool enable) = 0;

    //! Grabs a screenshot and saves it to file.
    virtual void ScreenShot(std::string filename) = 0;

    virtual void ClearStencil() = 0;

    virtual void PrepareStencil() = 0;

    virtual void StencilSetFirstShadowPass(shadow_mode mode) = 0;

    virtual void StencilSetSecondShadowPass(shadow_mode mode) = 0;

    virtual void StencilShadowRectangle(COLOUR colour) = 0;

};

#endif // GRAPHICS_SYSTEM_H
