// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef LANDSCAPE_H
#define LANDSCAPE_H

/*!
    \file landscape.h
    \brief File containing Landscape renderer class.
*/

#include "graphicssystem.h"

// A structure describing a rectangular region of terrain that should be updated.
// Updating is required after some terrain has been deformed or if it is being morphed.
struct update_rect
{
    int left;
    int top;
    int right;
    int bottom;
};

//! Landscape renderer.
class Landscape
{
public:

    //! Constructor. Loads the landscape from files.
    /*!
        Note that igs must point to a valid Rendering interface which should not be deconstructed
        before this object.
    */
    Landscape(std::string heightfield_filename, std::string texture_filename,
        std::string detail_texture_filename, std::string outer_texture_filename, IGraphicsSystem* igs);

    //! Destructor
    ~Landscape();

    //! Update function. Updates the LOD for the landscape.
    /*!
        You should call this no more than once per frame.
    */
    void Update(vector3 camera);

    //! Render function. Call this once per frame.
    /*!
        Renders the landscape.
    */
    void Render(vector3 camera);

    //! Returns a pointer to the heightfield.
    /*!
        The heightfield should be used by other parts of the program for collision detection.
        To deform terrain simply modify the terrain using this then call the update function below.
    */
    float* GetHeightfieldF() { return heightfield; }

    //! Terrain Update.
    /*!
        Call this function after modifying the terrain to recalculate the quadtree lookup-table and
        to modify the float version of the heightfield.
        The rectange is used to define the region of terrain that has been modified and needs
        to be updated.
    */
    void UpdateTerrainRectangle(update_rect& rect);

    //! Terrain texture update
    /*!
        Call this function to modify the terrain texture.
        width and height are the width are size of the buffer in pixels.
        x and y are is the top left corner of the region the buffer should modify on the map.
        tex_buffer is an ARGB pixel buffer of size width x height x 4 bytes.
    */
    void UpdateTerrainTexture(int width, int height, int x, int y, unsigned char *tex_buffer);

    //! Returns the height of the heightfield at a certain point. Performs interpolation if necessary.
    float GetHeight(float x, float y);

private:

    // The heightfield. A 513 x 513 array of floats.
    float* heightfield;

    // The "quadtree". A kind of array used to store precomputed error metrics.
    unsigned char* precalcerror;

    // A pointer to the rendering interface.
    IGraphicsSystem* IGS;
    
    // A pointer to the copy of the 2048x2048 texture in memory
    unsigned char* image;

    // All the 256 x 256 textures in the landscape got from the 
    // giant 2048 x 2048 texture.
    TEXTURE_HANDLE textures[64];
    
    // A single 256x256 texture used to disguise the low detail of the above
    // textures
    TEXTURE_HANDLE detail_texture;
    
    // A texture used to draw the lanscape surface when it's outside the (0,0)-(513,513) rectangle.
    TEXTURE_HANDLE outer_texture;

    // 513 x 513 Array of bools specifying which vertices are enabled.
    bool *vertexfield;

    // Enables a vertex. This vertex will then be used when rendering.
    inline void enablevertex(int x,int y)   {   *(vertexfield + x + y * 513)=true;  }

    // Checks whether or not a vertex is enabled.
    inline bool isvertexenabled(int x,int y)    {   return *(vertexfield + x + y * 513);    }

    // Given an x,y pair representing 1 of the 256 blocks in the landscape, and the square within
    // that block (left, right, bottom, top) this function will determine from the array of
    // precalculated errors the errors for the 4 subsquares in that square.
    void geterrors(int x, int y, int left, int top, int right, int bottom,
        int& nw_error, int& ne_error, int& sw_error, int& se_error);

    // Given an x,y pair representing 1 of the 256 blocks in the landscape, and the square within
    // that block (left, right, bottom, top) this function will enable 5 of the vertices within that
    // square, namely the centre and the 4 corners.
    void minimal_update(int x, int y, int left,int top, int right, int bottom);

    // Given an x,y pair representing 1 of the 256 blocks in the landscape, and the square within
    // that block (left, right, bottom, top) this function will recursively enable some of the
    // vertices of its subsquares in an attempt to get rid of some T-junctions which cause gaps
    // between polygons.
    void dependency_update(int x, int y, int left, int top, int right, int bottom);

    // This function calculates the precalculated error of each of its subsquares and divides that
    // by the distance from the camera and selectively updates those subsquares. This funtion is
    // called by the function update_block.
    void recursive_update(int x,int y,int left,int top, int right, int bottom);

    // Given an x,y pair representing 1 of the 256 blocks in the landscape this square will call the
    // recursive_update function (above) appropriately.
    void update_block(int x, int y);

    // Add options to change this from within the game.
    int geometric_detail;

    int local_detail;

    // Vertices in 1 of the 256 blocks
    VERTEX_XYZ_UV1_UV2 Vertices[33*33];

    // Indices of the vertices in 1 of the 256 blocks.
    unsigned short vertextable[33*33];

    // Array used to store an index list. This is used in a call to
    // DrawIndexedPrimitive( or equivalent OpenGL function )
    unsigned short Indices[8000];

    // Number of vertices in Vertices
    int vertexcount;

    // Number of indices in Indices
    int indexcount;

    // Fills the index list with the indices, ready for rendering
    void recursive_render(int x,int y,int left,int top, int right, int bottom);

    // Draws 1 of the 256 blocks in the landscape.
    void draw_block(int i,int j);

    // The miny and maxy of the AABBs of the 256 blocks.
    float block_miny[256];
    float block_maxy[256];

    // Vertex buffers for landscape segments
    VERTEX_BUFFER_HANDLE segments_vertices[256 * 4];
    int segments_vertices_count[256 * 4];

    // Indexbuffers for landscape segments
    INDEX_BUFFER_HANDLE segments_indices[256 * 4];
    int segments_indices_count[256 * 4];

    // Indicates what geomipmap to use for each block.
    int block_detail[256];

    // Indicates whether or not the AABB for a terrain block is in the frustum.
    bool block_visible[256];

    // Indicates which vertices are enabled at all 4 geomipmap levels.
    bool *vertexfield_levels;

    // Draws some polygon's to hide the gaps between adjacent blocks of different geomipmap levels
    // x1,y1 should be the one that the texture is currently set to.
    void draw_edge_strip(int x1, int y1, int x2, int y2);

};

#endif // LANDSCAPE_H
