// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "model.h"
#include "transform.h"
#include <iostream>

extern int g_stencil;

ANIMATION_INSTANCE_HANDLE IModel::CreateAnimationInstance()
{
    ANIMATION_INSTANCE_HANDLE handle = new float[GetVertices().size() * 6];

    return handle;
}

void IModel::DeleteAnimationInstance(ANIMATION_INSTANCE_HANDLE instance)
{
    delete[] instance;
}

void IModel::build_connectivity_list()
{
    if (g_stencil == 0) return;

    vertices = GetVertices();
    indices = GetIndices();

    faces.resize(indices.size() / 3);
    unsigned int f;
    for (f = 0; f < faces.size(); f++)
    {
        faces[f].vertex_indices[0] = indices[f * 3 + 0];
        faces[f].vertex_indices[1] = indices[f * 3 + 1];
        faces[f].vertex_indices[2] = indices[f * 3 + 2];
    }

    for (f = 0; f < faces.size(); f++)
    {
        for (int edge = 0; edge < 3; edge++)
        {
            faces[f].neighbour_indices[edge] = -1;
        }
    }

    //  for each face (A) in the object
    //      for each edge in A
    //          if we don't know this edges neighbour yet
    //              for each face (B) in the object (except A)
    //                  for each edge in B
    //                      if A's edge is the same as B's edge, then they are neighbouring each other on that edge
    //                          set the neighbour property for each face A and B, then move onto next edge in A

    unsigned int face_a, face_b, edge_a, edge_b;
    for (face_a = 0; face_a < faces.size(); face_a++)
    {
        for (edge_a = 0; edge_a < 3; edge_a++)
        {
            bool edge_found = false;

            if (faces[face_a].neighbour_indices[edge_a] == -1)
            {
                for (face_b = face_a + 1; face_b < faces.size(); face_b++)
                {
                    for (edge_b = 0; edge_b < 3; edge_b++)
                    {
                        if (faces[face_b].neighbour_indices[edge_b] == -1)
                        {
/*
                            if ((faces[face_a].vertex_indices[edge_a] == faces[face_b].vertex_indices[edge_b]) &&
                                (faces[face_a].vertex_indices[(edge_a+1) % 3] == faces[face_b].vertex_indices[(edge_b+1) % 3]) ||
                                (faces[face_a].vertex_indices[edge_a] == faces[face_b].vertex_indices[(edge_b+1) % 3]) &&
                                (faces[face_a].vertex_indices[(edge_a+1) % 3] == faces[face_b].vertex_indices[edge_b]))
                            {
                                faces[face_a].neighbour_indices[edge_a] = face_b;
                                faces[face_b].neighbour_indices[edge_b] = face_a;
                                edge_found = true;
                                break;
                            }
*/
                            vector3 vertex_a1 = vertices[faces[face_a].vertex_indices[edge_a]];
                            vector3 vertex_a2 = vertices[faces[face_a].vertex_indices[(edge_a+1) % 3]];

                            vector3 vertex_b1 = vertices[faces[face_b].vertex_indices[edge_b]];
                            vector3 vertex_b2 = vertices[faces[face_b].vertex_indices[(edge_b+1) % 3]];

                            if (((vertex_a1 == vertex_b1) && (vertex_a2 == vertex_b2)) ||
                                ((vertex_a1 == vertex_b2) && (vertex_a2 == vertex_b1)))
                            {
                                faces[face_a].neighbour_indices[edge_a] = face_b;
                                faces[face_b].neighbour_indices[edge_b] = face_a;
                                edge_found = true;
                                break;
                            }
                        }
                        if (edge_found) break;
                    }
                }
            }
        }
    }
}

vector3 project(const vector3& v, const vector3& light_vector_div_denom, const vector3 &plane_normal, const float& d)
{
    float numer = DotProduct(plane_normal, v) + d;

    return v - numer * light_vector_div_denom;
}

void IModel::cast_shadow(matrix44 object_to_world_transform, vector3 light_vector,
    ANIMATION_INSTANCE_HANDLE instance, shadow_mode mode, float clip_y)
{
    unsigned int f;
    vector3 light_vec = light_vector.normalize();
    vector3 light_vec_reverse = reverse_rotate(light_vec, object_to_world_transform).normalize();


    if (instance != NULL)
    {
        for (unsigned int v = 0; v < vertices.size(); v++)
        {
            vertices[v].x = instance[v * 6 + 0];
            vertices[v].y = instance[v * 6 + 1];
            vertices[v].z = instance[v * 6 + 2];

            // Normals are ignored here. instance[v * 6 + 3] is normal.x

        }
    }

    // Determine which faces are visible by the light
    for (f = 0; f < faces.size(); f++)
    {
        // Determine the normal of each face
        vector3 origin = vertices[faces[f].vertex_indices[0]];
        vector3 v1 = vertices[faces[f].vertex_indices[1]] - origin;
        vector3 v2 = vertices[faces[f].vertex_indices[2]] - origin;

        vector3 normal = CrossProduct(v1, v2); //.normalize();

        // Determine the angle between this and the light vector
        float dot = DotProduct(normal, light_vec_reverse);

        if (dot > 0.0f)
        {
            faces[f].visible = false;
        }
        else
        {
            faces[f].visible = true;
        }
    }

    vector3 plane1 = reverse_transform(vector3(0.0f, clip_y, 0.0f), object_to_world_transform);
    vector3 plane2 = reverse_transform(vector3(100.0f, clip_y, 100.0f), object_to_world_transform);
    vector3 plane3 = reverse_transform(vector3(100.0f, clip_y, 0.0f), object_to_world_transform);
    vector3 plane_origin = plane1;
    vector3 plane_normal = CrossProduct(plane2 - plane1, plane3 - plane1);
    plane_normal.normalize();

    float d = -DotProduct(plane_normal, plane_origin);
    float denom = DotProduct(plane_normal, light_vec_reverse);
    vector3 light_vec_reverse_div_denom = (1.0f / denom) * light_vec_reverse;


    VERTEX_XYZ points[10000];
    int number_of_points = 0;

    if (mode == Z_FAIL_SHADOWS)
    {
        for (f = 0; f < faces.size(); f++)
        {
            if (faces[f].visible)
            {
                vector3 v[6];

                v[0] = vertices[faces[f].vertex_indices[0]];
                v[1] = vertices[faces[f].vertex_indices[1]];
                v[2] = vertices[faces[f].vertex_indices[2]];

                v[3] = project(v[0], light_vec_reverse_div_denom, plane_normal, d);
                v[4] = project(v[1], light_vec_reverse_div_denom, plane_normal, d);
                v[5] = project(v[2], light_vec_reverse_div_denom, plane_normal, d);

                points[number_of_points + 0].point = v[3];
                points[number_of_points + 1].point = v[5];
                points[number_of_points + 2].point = v[4];

                number_of_points += 3;

                points[number_of_points + 0].point = v[0];
                points[number_of_points + 1].point = v[1];
                points[number_of_points + 2].point = v[2];

                number_of_points += 3;

                for (int edge = 0; edge < 3; edge++)
                {
                    int neighbour = faces[f].neighbour_indices[edge];

                    // If its neighbouring face is not visible, or there is no neighbour then this edge casts a shadow
                    if ((faces[neighbour].visible == false) || (neighbour == -1))
                    {
                        // Get the points on the edge
                        vector3 v1 = v[edge];
                        vector3 v2 = v[(edge+1) % 3];

                        // Get the two vertices in the distance
                        vector3 v3 = v[edge + 3];
                        vector3 v4 = v[((edge+1) % 3) + 3];

                        points[number_of_points + 0].point = v1;
                        points[number_of_points + 1].point = v3;
                        points[number_of_points + 2].point = v2;

                        points[number_of_points + 3].point = v3;
                        points[number_of_points + 4].point = v4;
                        points[number_of_points + 5].point = v2;

                        number_of_points += 6;
                    }
                }
            }
        }
    }
    else // (mode == Z_PASS_SHADOWS)
    {
        for (f = 0; f < faces.size(); f++)
        {
            if (faces[f].visible)
            {
                for (int edge = 0; edge < 3; edge++)
                {
                    int neighbour = faces[f].neighbour_indices[edge];

                    // If its neighbouring face is not visible, or there is no neighbour then this edge casts a shadow
                    if ((faces[neighbour].visible == false) || (neighbour == -1))
                    {
                        // Get the points on the edge
                        vector3 v1 = vertices[faces[f].vertex_indices[edge]];
                        vector3 v2 = vertices[faces[f].vertex_indices[(edge+1) % 3]];

                        // Get the two vertices in the distance
                        vector3 v3 = project(v1, light_vec_reverse_div_denom, plane_normal, d);
                        vector3 v4 = project(v2, light_vec_reverse_div_denom, plane_normal, d);

                        points[number_of_points + 0].point = v1;
                        points[number_of_points + 1].point = v3;
                        points[number_of_points + 2].point = v2;

                        points[number_of_points + 3].point = v3;
                        points[number_of_points + 4].point = v4;
                        points[number_of_points + 5].point = v2;

                        number_of_points += 6;
                    }
                }
            }
        }
    }

    IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, object_to_world_transform);

    IGS->StencilSetFirstShadowPass(mode);
    IGS->DrawTrianglesUP(TRIANGLE_LIST, VTX_XYZ, (float*)points, number_of_points);

    IGS->StencilSetSecondShadowPass(mode);
    IGS->DrawTrianglesUP(TRIANGLE_LIST, VTX_XYZ, (float*)points, number_of_points);
}

