// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "SDL.h"
#include "SDL_keysym.h"
#include <iostream>
#include <sstream>
#include <stdlib.h>
#include "application.h"
#include "keyboard.h"
#include "filenames.h"

int g_width = 640;
int g_height= 480;
int g_stencil = 8;
int g_landscape_detail = 4000;
int g_texture = 1;

int main(int argc, char* argv[])
{
    // Information about the current video settings
    const SDL_VideoInfo* info = NULL;

    // Colour depth in bits of our window
    int bpp = 32;
    // Flags we will pass into SDL_SetVideoMode
    int flags = 0;

    // First initialise SDL's video subsystem
    if (SDL_Init(SDL_INIT_VIDEO) < 0)
    {
        std::cout << "Video initialisation failed: " << SDL_GetError() << std::endl;
        exit(1);
    }
    
    // Shutdown SDL at exit
    atexit(SDL_Quit);

    // Let's get some video information
    info = SDL_GetVideoInfo();
    if (!info)
    {
        std::cout << "Video Query failed: " << SDL_GetError() << std::endl;
        exit(1);
    }
    int current_bpp = info->vfmt->BitsPerPixel;

    // Call a function in filename.h
    set_argv_zero(std::string(argv[0]));

    for(int i = 1; i<argc; i++)
    {
        if(argv[i][0]=='-') switch(argv[i][1])
        {
            case 'w': g_width   = atoi(&argv[i][2]); break;
            case 'h': g_height  = atoi(&argv[i][2]); break;
            case 'b': bpp       = atoi(&argv[i][2]); break;
            case 's': g_stencil = atoi(&argv[i][2]); break;
            case 'l': g_landscape_detail = atoi(&argv[i][2]); break;
            case 't': g_texture = atoi(&argv[i][2]); break;
            default: std::cout << "unknown commandline option" << std::endl;
        }
        else std::cout << "unknown commandline argument" << std::endl;
    };

    // Now we want to setup our requested window attributes for out OpenGL
    // window. We want *at least* 5 bits of red, green and blue. We also
    // want at least a 15-bit depth buffer.
    //
    // The last thing we do is request a double buffered window. '1' turns
    // on double buffering '0' turns it off.
    //
    // Note that we do not use SDL_DOUBLEBUF in the flags to
    // SDL_SetVideoMode. That does not affect the GL attribute state only
    // the standard 2D blitting setup.

    SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 15);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, g_stencil);

    // We want to request that SDL provide us with an OpenGL window in a
    // fullscreen mode.

    flags = SDL_OPENGL | SDL_FULLSCREEN;

    // Set the video mode
    if (SDL_SetVideoMode(g_width, g_height, bpp, flags) == 0)
    {
        // Failed. Try again with current bpp.
        if (SDL_SetVideoMode(g_width, g_height, current_bpp, flags) == 0)
        {
            // Failed. Try again with no stencil buffer
            SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, g_stencil = 0);
            if (SDL_SetVideoMode(g_width, g_height, current_bpp, flags) == 0)
            {
                std::cout << "Failed to set video mode: " << SDL_GetError() << std::endl;
                exit(1);
            }
        }
    }

    SDL_GL_GetAttribute(SDL_GL_STENCIL_SIZE, &g_stencil);

    g_width = SDL_GetVideoSurface()->w;
    g_height= SDL_GetVideoSurface()->h;

    // Set the window caption
    SDL_WM_SetCaption("Slickworm", "Slickworm");

    // Hide the mouse cursor
    SDL_ShowCursor(0);

    // Grab the mouse and keyboard.
    SDL_WM_GrabInput(SDL_GRAB_ON);

/*
    // Enable Unicode translation
    SDL_EnableUNICODE( 1 );
*/

    Application App;


    bool exit = false;
    while(!exit)
    {
        // Call the application's update function
        App.Update();

        // Poll for keypresses
        SDL_Event event;
        while (SDL_PollEvent(&event))
        {
            // We are only worried about SDL_KEYDOWN events
            switch(event.type)
            {
                case SDL_KEYDOWN:
                case SDL_KEYUP:
                        if (event.key.keysym.sym == SDLK_ESCAPE) exit = 1;
                        Keyboard::Instance().KeyEventNotify(&event.key);
                    break;

                case SDL_QUIT:
                    exit = true;
                    break;

                default:
                    break;
            }
        }

        if (Keyboard::Instance().KeyPressed(KEY_ESC)) exit = 1;
    }

    // Because SDL_Quit was passed as a parameter to atexit I don't need to call it here.
    return 0;
}
